from meiga import Result

from petisco.extra.fastapi.controller.fastapi_default_response import (
    FASTAPI_DEFAULT_RESPONSE,
)


def fastapi_success_handler(result: Result):
    try:
        response = (
            FASTAPI_DEFAULT_RESPONSE
            if result.value is True
            else {"result": result.value}
        )
    except Exception:
        response = FASTAPI_DEFAULT_RESPONSE

    return response



from typing import List

from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.application.notifier.not_implemented_notifier import (
    NotImplementedNotifier,
)
from petisco.base.misc.builder import Builder


def get_default_notifier_dependencies() -> List[Dependency]:

    return [
        Dependency(
            name="notifier",
            default_builder=Builder(NotImplementedNotifier),
            envar_modifier="PETISCO_NOTIFIER_TYPE",
        )
    ]


def get_slack_notifier_dependencies(token: str, channel: str) -> List[Dependency]:
    from petisco.extra.slack.application.notifier.slack_notifier import SlackNotifier

    return [
        Dependency(
            name="notifier",
            default_builder=Builder(SlackNotifier, token=token, channel=channel),
            envar_modifier="PETISCO_NOTIFIER_TYPE",
            builders={"not_implemented": Builder(NotImplementedNotifier)},
        )
    ]



from typing import Callable


class MessageHandlerReturnsNoneError(ValueError):
    def __init__(self, handler: Callable):
        handler_name = getattr(handler, "__name__", repr(handler))
        handler_module = getattr(handler, "__module__") + "."
        message = f"Subscribers (event handlers) must return a Result object. Please check {handler_module}{handler_name}"
        super().__init__(message)



from typing import Callable, List, Type

from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.message_subscriber import MessageSubscriber


class NotImplementedMessageConsumer(MessageConsumer):
    def start(self):
        pass

    def add_subscribers(self, subscribers: List[MessageSubscriber]):
        pass

    def add_subscriber_on_dead_letter(self, subscriber: Type[MessageSubscriber]):
        pass

    def add_handler_on_store(self, handler: Callable):
        pass

    def add_subscriber_on_queue(self, queue_name: str, handler: Callable):
        pass

    def unsubscribe_subscriber_on_queue(self, queue_name: str):
        pass

    def resume_subscriber_on_queue(self, queue_name: str):
        pass

    def stop(self):
        pass



from abc import ABC
from typing import Any, Callable, Dict

from meiga import Result


def default_failure_handler(result: Result, error_map: Dict):
    error_type = type(result.value)
    mapped_result = error_map.get(error_type, result)
    return mapped_result


class ResultMapper(ABC):
    def __init__(
        self,
        error_map: Dict[type, Any] = None,
        success_handler: Callable = lambda result: result,
        failure_handler: Callable = default_failure_handler,
    ):
        self.error_map = error_map if error_map is not None else dict()
        self.success_handler = success_handler
        self.failure_handler = failure_handler

    def map(self, result: Result):
        if result.is_success:
            return self.success_handler(result)
        else:
            return self.failure_handler(result, self.error_map)



import copy
from abc import ABC, abstractmethod
from typing import Dict, List

from petisco.base.domain.message.message import Message


class MessageBus(ABC):
    @classmethod
    def __repr__(cls):
        return cls.__name__

    @classmethod
    def info(cls) -> Dict:
        return {"name": cls.__name__}

    def _set_additional_meta(self, meta: Dict):
        self.additional_meta = meta

    def with_meta(self, meta: Dict):
        if not isinstance(meta, Dict):
            raise TypeError("MessageBus.with_meta() expect a dict")
        event_bus = copy.copy(self)
        event_bus._set_additional_meta(meta)
        return event_bus

    def get_configured_meta(self) -> Dict:
        configured_meta = {}
        if hasattr(self, "additional_meta"):
            configured_meta = {**configured_meta, **self.additional_meta}
        return configured_meta

    @abstractmethod
    def publish(self, message: Message):
        raise NotImplementedError

    @abstractmethod
    def retry_publish_only_on_store_queue(self, message: Message):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError

    def _check_is_message(self, message: Message):
        if not message or not issubclass(message.__class__, Message):
            raise TypeError("MessageBus only publishes Message objects")

    def publish_list(self, messages: List[Message]):
        for message in messages:
            self.publish(message)

    def retry_publish_list_only_on_store_queue(self, messages: List[Message]):
        for message in messages:
            self.retry_publish_only_on_store_queue(message)



from abc import abstractmethod
from typing import Generic, List, TypeVar

from meiga import (
    BoolResult,
    Error,
    Failure,
    NotImplementedMethodError,
    Result,
    Success,
    isSuccess,
)
from meiga.decorators import meiga
from sqlalchemy import create_engine
from sqlmodel import Session, SQLModel, select

from petisco import Uuid
from petisco.base.application.patterns.crud_repository import (
    AggregateRootType,
    CrudRepository,
)
from petisco.base.domain.errors.defaults.already_exists import (
    AggregateAlreadyExistError,
)
from petisco.base.domain.errors.defaults.not_found import AggregateNotFoundError

engine = create_engine("sqlite:///database.db", echo=True)
SQLModel.metadata.create_all(engine)

SQLModelType = TypeVar("SQLModelType", bound=SQLModel)


class SQLModelCrudRepository(
    Generic[SQLModelType, AggregateRootType], CrudRepository[AggregateRootType]
):
    @abstractmethod
    def get_aggregate_root(
        self, sql_model: SQLModel
    ) -> Result[AggregateRootType, Error]:
        return NotImplementedMethodError

    @abstractmethod
    def get_sql_model(
        self, aggregate_root: AggregateRootType
    ) -> Result[SQLModel, Error]:
        return NotImplementedMethodError

    @abstractmethod
    def get_sql_model_type(self) -> SQLModelType:
        return NotImplementedMethodError

    @meiga
    def save(self, aggregate_root: AggregateRootType) -> BoolResult:

        with Session(engine) as session:
            model = self.get_sql_model_type()
            statement = select(model).where(
                model.aggregate_id == aggregate_root.aggregate_id.value
            )
            sql_model = session.exec(statement).first()
            if sql_model:
                return Failure(AggregateAlreadyExistError(aggregate_root.aggregate_id))

            sql_model = self.get_sql_model(aggregate_root).unwrap_or_return()

            session.add(sql_model)
            session.commit()

            return isSuccess

    @meiga
    def retrieve(self, aggregate_id: Uuid) -> Result[AggregateRootType, Error]:
        with Session(engine) as session:
            model = self.get_sql_model_type()
            statement = select(model).where(model.aggregate_id == aggregate_id.value)
            sql_model = session.exec(statement).first()
            if sql_model is None:
                return Failure(AggregateNotFoundError(aggregate_id))
            aggregate_root = self.get_aggregate_root(sql_model).unwrap_or_return()
            return Success(aggregate_root)

    def update(self, aggregate_root: AggregateRootType) -> BoolResult:
        with Session(engine) as session:
            model = self.get_sql_model_type()
            statement = select(model).where(
                model.aggregate_id == aggregate_root.aggregate_id.value
            )
            sql_model = session.exec(statement).first()
            if sql_model is None:
                return Failure(AggregateNotFoundError(aggregate_root.aggregate_id))

            sql_model = self.get_sql_model(aggregate_root)
            session.add(sql_model)
            session.commit()
            return isSuccess

    def remove(self, aggregate_id: Uuid) -> BoolResult:
        with Session(engine) as session:
            model = self.get_sql_model_type()
            statement = select(model).where(model.aggregate_id == aggregate_id.value)
            sql_model = session.exec(statement).first()
            if sql_model is None:
                return Failure(AggregateNotFoundError(aggregate_id))

            session.delete(sql_model)
            session.commit()

            return isSuccess

    def retrieve_all(self) -> Result[List[AggregateRootType], Error]:
        with Session(engine) as session:
            model = self.get_sql_model_type()
            statement = select(model)
            sql_models = session.exec(statement)

            all = [
                self.get_aggregate_root(sql_model).unwrap_or_return()
                for sql_model in sql_models
            ]
            return Success(all)



import sys
import traceback
from typing import Any, Dict, List, Optional

from meiga import Error


class UnknownError(Error):
    def __init__(
        self,
        exception: Exception,
        input_parameters=None,
        executor=None,
        traceback=None,
        filename=None,
        lineno=None,
        filter_parameters: List[str] = None,
        meta: Dict = None,
    ):
        self.message = f"{exception.__class__.__name__}: {str(exception)}"
        self.input_parameters = self._sanitize_input_params(input_parameters)
        self._filter_input_parameters(filter_parameters)
        self.exception = exception
        self.executor = executor
        self.traceback = traceback
        self.filename = filename
        self.lineno = lineno
        self.meta = meta if meta else {}

    def _sanitize_input_params(self, input_parameters):
        if isinstance(input_parameters, tuple):
            return {
                f"param_{i + 1}": param if not isinstance(param, bytes) else "bytes"
                for i, param in enumerate(input_parameters)
            }
        elif isinstance(input_parameters, dict):
            return {
                k: v if not isinstance(v, bytes) else "bytes"
                for k, v in input_parameters.items()
            }
        else:
            return None

    def _filter_input_parameters(self, filter_parameters: List[str] = None):
        if filter_parameters:
            self.input_parameters = {
                k: v
                for k, v in self.input_parameters.items()
                if k not in filter_parameters
            }

    def __repr__(self):
        executor_str = f" ({self.executor})" if self.executor else ""
        traceback_str = f"\n{self.traceback}" if self.traceback else ""
        input_parameters_str = (
            f"\nInput Parameters: {str(self.input_parameters)}"
            if self.input_parameters
            else ""
        )
        filename_str = f"\n{self.filename}" if self.filename else ""
        lineno_str = f"\n{self.lineno}" if self.lineno else ""
        return f"{self.__class__.__name__}{executor_str}: {self.message}.{traceback_str}.{input_parameters_str}{filename_str}{lineno_str}"

    @classmethod
    def from_exception(
        cls,
        exception: Exception,
        arguments: Any,
        class_name: Optional[str] = None,
    ):
        _, _, tb = sys.exc_info()
        tb = traceback.extract_tb(tb)[-1]
        filename = tb.filename if tb and hasattr(tb, "filename") else None
        lineno = tb.lineno if tb and hasattr(tb, "lineno") else None

        unknown_error = cls(
            exception=exception,
            input_parameters=arguments,
            executor=class_name,
            traceback=traceback.format_exc(),
            filename=filename,
            lineno=lineno,
        )
        return unknown_error



from petisco.base.domain.errors.unknown_error import UnknownError


class NotifierExceptionMessage:
    title: str
    executor: str
    exception: Exception
    traceback: str
    filename: str = None
    lineno: str = None
    input_parameters: dict = None
    meta: dict = None

    def __init__(
        self,
        title: str,
        executor: str,
        exception: Exception,
        traceback: str,
        filename: str = None,
        lineno: str = None,
        input_parameters: dict = None,
        meta: dict = None,
    ):
        self.title = title
        self.executor = executor
        self.exception = exception
        self.traceback = traceback
        self.filename = filename
        self.lineno = lineno
        self.input_parameters = input_parameters
        self.meta = meta

    def update_meta(self, meta: dict):
        self.meta = {**self.meta, **meta}

    @staticmethod
    def from_unknown_error(unknown_error: UnknownError, title: str):
        return NotifierExceptionMessage(
            title=title,
            executor=unknown_error.executor,
            exception=unknown_error.exception,
            traceback=unknown_error.traceback,
            filename=unknown_error.filename,
            lineno=unknown_error.lineno,
            input_parameters=unknown_error.input_parameters,
            meta=unknown_error.meta,
        )



import os
from functools import wraps
from inspect import signature

import elasticapm
from meiga import Error, Failure
from meiga.on_failure_exception import OnFailureException

from petisco.base.application.middleware.notifier_middleware import NotifierMiddleware
from petisco.base.application.middleware.print_middleware import PrintMiddleware
from petisco.base.domain.errors.unknown_error import UnknownError


def get_middleware_classes(config):
    def gettype(name):
        lookup_table = {
            "NotifierMiddleware": NotifierMiddleware,
            "PrintMiddleware": PrintMiddleware,
        }
        my_type = lookup_table.get(name)
        if not my_type:
            raise TypeError(
                f"Value {name} in PETISCO_DEFAULT_MIDDLEWARES is not valid. Please, use {list(lookup_table.keys())}"
            )
        return my_type

    middlewares_classes = getattr(config, "middlewares", [])
    if not middlewares_classes:
        default_middlewares_names = os.getenv("PETISCO_DEFAULT_MIDDLEWARES", [])
        if default_middlewares_names:
            default_middlewares_names = default_middlewares_names.split(",")
            middlewares_classes = [
                gettype(middleware_name)
                for middleware_name in default_middlewares_names
            ]

    return middlewares_classes


def wrapper(execute_func, wrapped_class_name, config, mapper):
    @wraps(execute_func)
    def wrapped(*args, **kwargs):
        middleware_classes = get_middleware_classes(config)
        arguments = signature(execute_func).bind(*args, **kwargs).arguments
        arguments.pop("self")
        middlewares = [
            klass(wrapped_class_name, arguments) for klass in middleware_classes
        ]

        for middleware in middlewares:
            middleware.before()

        try:
            result = execute_func(*args, **kwargs)
        except OnFailureException as exc:
            result = exc.result
        except Error as error:
            result = Failure(error)
        except Exception as exception:
            unknown_error = UnknownError.from_exception(
                exception=exception,
                arguments=arguments if len(arguments) > 0 else args,
                class_name=wrapped_class_name,
            )
            result = Failure(unknown_error)
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

        for middleware in middlewares:
            if result:
                middleware.after(result)

        mapped_result = mapper.map(result)

        return mapped_result

    return wrapped



import argparse
import os
from datetime import datetime

from petisco import Uuid, __version__


def has_args(args):
    is_active = False
    for arg in vars(args):
        is_active = is_active or getattr(args, arg)
    return is_active


def rename_template(original_name: str, replacement: str):

    for dname, _, _ in os.walk("."):
        if original_name in dname:
            os.rename(dname, dname.replace(original_name, replacement))

    blacklist = [".git/", ".idea"]
    for dname, dirs, files in os.walk("."):
        for fname in files:
            fpath = os.path.join(dname, fname)
            rewrite = True
            for blackfolder in blacklist:
                if blackfolder in fpath:
                    rewrite = False
            if rewrite is False:
                continue
            try:
                with open(fpath) as f:
                    s = f.read()
                s = s.replace(original_name, replacement)
                with open(fpath, "w") as f:
                    f.write(s)
            except Exception:
                pass


def main():
    parser = argparse.ArgumentParser(
        prog="petisco 🍪",
        description="petisco is a framework for helping Python developers to build clean Applications",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-v", "--version", action="store_true", help="show petisco version number."
    )
    parser.add_argument("-uuid", "--uuid", action="store_true", help="show an UUID v4.")
    parser.add_argument(
        "-utcnow", "--utcnow", action="store_true", help="show a utc now datetime"
    )
    parser.add_argument(
        "-rt",
        "--rename-template",
        action="store",
        dest="rename_template_replacement",
        default=None,
        help="Rename a petisco service template.",
    )
    parser.add_argument(
        "-tn",
        "--original-template-name",
        action="store",
        dest="original_template_name",
        default="taskmanager",
        help="Rename a petisco service template.",
    )

    args = parser.parse_args()

    if not has_args(args):
        parser.print_help()
    else:
        if args.version:
            print(f"petisco 🍪 => {__version__}")
            return

        if args.uuid:
            print(Uuid.v4().value)
            return

        if args.utcnow:
            print(datetime.utcnow())
            return

        if args.rename_template_replacement:
            print(
                f"petisco 🍪 => Changing {args.original_template_name} for {args.rename_template_replacement}..."
            )
            rename_template(
                args.original_template_name, args.rename_template_replacement
            )
            return



from abc import ABC, abstractmethod
from functools import wraps
from types import FunctionType
from typing import Any

import elasticapm
from meiga import Error, Failure, NotImplementedMethodError, Result
from meiga.on_failure_exception import OnFailureException

from petisco.base.application.use_case.use_case_uncontrolled_error import (
    UseCaseUncontrolledError,
)


def wrapper(method, wrapped_class_name):
    @wraps(method)
    def wrapped(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except OnFailureException as exception:
            return exception.result
        except Error as error:
            return Failure(error)
        except Exception as exception:
            uncontrolled_error = UseCaseUncontrolledError.from_exception(
                exception=exception, arguments=args, class_name=wrapped_class_name
            )
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

            return Failure(uncontrolled_error)

    return wrapped


class MetaUseCase(type, ABC):
    def __new__(mcs, name, bases, namespace):
        new_class_dict = {}
        if "execute" not in namespace:
            raise NotImplementedError(
                "Petisco UseCase must implement an execute method"
            )

        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "execute":
                attribute = wrapper(attribute, name)
            new_class_dict[attributeName] = attribute
        return type.__new__(mcs, name, bases, new_class_dict)

    @abstractmethod
    def execute(self, *args, **kwargs) -> Result[Any, Error]:
        return NotImplementedMethodError


class UseCase(metaclass=MetaUseCase):
    @abstractmethod
    def execute(self, *args, **kwargs) -> Result[Any, Error]:
        return NotImplementedMethodError



from abc import ABC, abstractmethod

from meiga import Result


class Middleware(ABC):
    def __init__(self, wrapped_class_name, wrapped_class_input_arguments):
        self.wrapped_class_name = wrapped_class_name
        self.wrapped_class_input_arguments = wrapped_class_input_arguments

    @abstractmethod
    def before(self):
        pass

    @abstractmethod
    def after(self, result: Result):
        pass
